""" This module contains the abstract class for cloud storage factory. """
from abc import ABC, abstractmethod


class CloudStorageFactory(ABC):
    """Abstract class for cloud storage factory."""

    @abstractmethod
    def create_client(self):
        """
        Create the cloud storage client.
        """
        pass

    @abstractmethod
    def upload(self, file_path, destination):
        """Upload data to the cloud storage.

        Parameters:
        - file_path: The path to the file to upload.
        - destination: The destination path in the cloud storage.
        """
        pass

    @abstractmethod
    def download(self, source, file_path):
        """Download data from the cloud storage.

        Parameters:
        - source: The source path in the cloud storage.
        - file_path: The path to save the downloaded file.
        """
        pass



""" S3 Cloud Storage"""
from typing import Optional
from unicloud.abstract_class import CloudStorageFactory
import boto3


class S3(CloudStorageFactory):
    """S3 Cloud Storage"""

    def __init__(
        self,
        aws_access_key_id: str,
        aws_secret_access_key: str,
        region_name: Optional[str] = None,
    ):
        """
        Initialize the AWS S3 client with credentials and region information.

        Parameters
        ----------
        aws_access_key_id: [str]
            Your AWS access key ID.
        aws_secret_access_key: [str]
            Your AWS secret access key.
        region_name: Optional[str]
            The name of the AWS region to connect to. Default is None.
        """
        self.aws_access_key_id = aws_access_key_id
        self.aws_secret_access_key = aws_secret_access_key
        self.region_name = region_name
        self._client = self.create_client()

    @property
    def client(self):
        """client."""
        return self._client

    def create_client(self):
        """
        Creates and returns an AWS S3 client instance.

        initializing the AWS S3 client, passing credentials directly is one option. Another approach is to use AWS
        IAM roles for EC2 instances or to configure the AWS CLI with aws configure, which sets up the credentials
        file used by boto3. This can be a more secure and manageable way to handle credentials, especially in
        production environments.

        Initialize the S3 client with AWS credentials and region.
        """

        return boto3.client(
            "s3",
            region_name=self.region_name,
            aws_access_key_id=self.aws_access_key_id,
            aws_secret_access_key=self.aws_secret_access_key,
        )

    def upload(self, path: str, bucket_path: str):
        """Upload a file to S3.

        Parameters
        ----------
        path: [str]
            The path to the file to upload.
        bucket_path: [str]
            The bucket_path in the format "bucket_name/object_name".
        """
        bucket_name, object_name = bucket_path.split("/", 1)
        self.client.upload_file(path, bucket_name, object_name)
        print(f"File {path} uploaded to {bucket_path}.")

    def download(self, bucket_path: str, file_path: str):
        """Download a file from S3.

        Parameters
        ----------
        bucket_path: [str]
            The bucket_path in the format "bucket_name/object_name".
        file_path: [str]
            The path to save the downloaded file.
        """
        bucket_name, object_name = bucket_path.split("/", 1)
        self.client.download_file(bucket_name, object_name, file_path)
        print(f"File {bucket_path} downloaded to {file_path}.")



"""unicloud.

Python package doing xyz for .com

https://gitlab.com/cquest1/tools/unicloud
"""
try:
    from importlib.metadata import PackageNotFoundError, version  # type: ignore
except ImportError:  # pragma: no cover
    from importlib_metadata import PackageNotFoundError, version  # type: ignore


try:
    __version__ = version(__name__)
except PackageNotFoundError:  # pragma: no cover
    __version__ = "unknown"

__author__ = "Mostafa Farrag"
__email__ = "mostafa.farrag@.com"



"""Authenticate."""
import base64
import json
import os

from google.cloud import storage
from google.oauth2 import service_account as service_account_gcp


def create_client(
    project_id: str,
    your_service_key: str = None,
) -> storage.client.Client:
    """create_client.

        the returned client deals with everything related to the specific project.

    Parameters
    ----------
    project_id: [str]
        project id
    your_service_key: [string]
        path to your saved api-key json file

    Returns
    -------
    google cloud storage client object
    """
    if "GOOGLE_APPLICATION_CREDENTIALS" in os.environ:
        credentials = service_account_gcp.Credentials.from_service_account_file(
            os.environ["GOOGLE_APPLICATION_CREDENTIALS"]
        )
        client = storage.Client(project=project_id, credentials=credentials)
    elif "EE_PRIVATE_KEY" in os.environ and "EE_SERVICE_ACCOUNT" in os.environ:
        # key need to be decoded into a dict/json object
        service_key_content = base64.b64decode(
            eval(f"b'{os.environ['EE_PRIVATE_KEY']}'")
        ).decode()

        service_key_content = json.loads(service_key_content)
        # connection to the service account
        client = storage.Client.from_service_account_info(service_key_content)
    elif your_service_key:
        credentials = service_account_gcp.Credentials.from_service_account_file(
            your_service_key
        )
        client = storage.Client(project=project_id, credentials=credentials)
    else:
        raise ValueError(
            "Since the GOOGLE_APPLICATION_CREDENTIALS and the EE_PRIVATE_KEY and EE_SERVICE_ACCOUNT are not in your env variables you have to provide a path to your service account"
        )

    return client



"""gsutils."""
import os

from google.cloud import storage


def is_gs_uri(uri: str) -> bool:
    """Check whether the URI points to a Google Cloud Storage object."""
    return uri.startswith("gs://")


def parse_gs_uri(uri):
    """Parse the Google Cloud Storage URI into bucket and file name."""
    rel_path = uri.removeprefix("gs://")
    bucket_id = rel_path.split("/")[0]
    file_name = rel_path.removeprefix(bucket_id)[1:]
    return bucket_id, file_name


def file_exists(uri: str, gcp_client: storage.Client) -> bool:
    """Check whether a file exists locally or in a GCP bucket."""
    if is_gs_uri(uri):
        bucket_id, file_name = parse_gs_uri(uri)
        bucket = gcp_client.get_bucket(bucket_id)
        return bucket.get_blob(file_name) is not None
    else:
        return os.path.exists(uri)


def open_file(uri: str, mode: str, gcp_client: storage.Client):
    """Open a local or remote file."""
    if is_gs_uri(uri):
        bucket_id, file_name = parse_gs_uri(uri)
        bucket = gcp_client.get_bucket(bucket_id)
        blob = bucket.blob(file_name)
        outfile = blob.open(mode)
    else:
        outfile = open(uri, mode)
    return outfile



"""unicloud.

Python package doing xyz for .com

https://gitlab.com/cquest1/tools/unicloud
"""


def main():
    """Execute unicloud functionalities."""
    return True


if __name__ == "__main__":
    main()



"""Bucket related functions."""
import io
import shutil
from typing import List, Union

import pandas as pd
from google.cloud import storage
from loguru import logger

from unicloud.gs import open_file


def get_bucket(
    client: storage.client.Client, bucket: str, project_id: str
) -> storage.bucket.Bucket:
    """get_bucket.

        get_bucket returns the bucket object

    Parameters
    ----------
    client : [google.cloud.storage.client.Client]
        google client object
    bucket : [str]
        bucket id
    project_id : [str]
        project name

    Returns
    -------
    storage.bucket.Bucket

    Examples
    --------
    >>> from unicloud.authenticate import create_client
    >>> your_service_key = "/service-key-gee-data-access.json"
    >>> client = create_client(your_service_key, project_id)
    >>> bucket = "-datasets"
    >>> project_id = "gee-data-access"
    >>> bucket_usr = get_bucket(client, bucket, project_id)
    """
    return storage.Bucket(client, bucket, user_project=project_id)


def get_bucket_list(client_usr: storage.client.Client) -> List[str]:
    """get_bucket_list.

        get_bucket_list returns a list of all the content of a bucket

    Parameters
    ----------
    client_usr: [google.cloud.storage.client.Client]
        google client object

    Returns
    -------
    bucket_list: [list]
        list of the names of the files exist in the bucket as a string

    Examples
    --------
    >>> from google.cloud import storage
    >>> project_id = "gee-data-access"
    >>> credentials = ""
    >>> client = storage.Client(project=project_id, credentials=credentials)
    >>> bucket_content = get_bucket_list(client)
    """
    buckets = list(client_usr.list_buckets())
    bucket_list = []
    for bucketi in buckets:
        # logger.debugprint(bucket.id)
        bucket_list.append(bucketi.id)
    return bucket_list


def get_bucket_contents(
    client_usr: storage.client.Client, bucket_id: str, project_id: str
) -> List[str]:
    """get_bucket_contents.

        get_bucket_contents returns the names of the buckets exists inside a project

    Parameters
    ----------
    client_usr: []

    bucket_id : [str]
        bucket id
    project_id: [str]
        project id
    Returns
    -------
    content : [list]
        list of the names of the buckets exist in the project as a string

    Examples
    --------
        >>> from google.cloud import storage
        >>> client = storage.Client(project=project_id, credentials=credentials)
        >>> bucket_content = get_bucket_list(client)
    """
    bucket_usr = storage.Bucket(client_usr, bucket_id, user_project=project_id)
    all_blobs = list(client_usr.list_blobs(bucket_usr))
    content = []
    for blobi in all_blobs:
        content.append(blobi.path.split("/")[-1])
    return content


def get_file_from_bucket(
    file: str, usr_bucket: storage.bucket.Bucket
) -> storage.blob.Blob:
    """get_file_from_bucket.

        get_file_from_bucket retrives a file from a given bucket object as a blob object

    Parameters
    ----------
    file: [str]
        file name you want to retrive from the bucket
    usr_bucket: [google.cloud.storage.bucket.Bucket]
        the bucket object

    Returns
    -------
    blob : [google.cloud.storage.blob.Blob]
        binary large object
    """
    # get the file from the bucket
    blob_usr = usr_bucket.get_blob(file)
    # logger.debug(blob.download_as_bytes())
    if blob_usr is None:
        raise ValueError(f"The {file} does not exist in the {usr_bucket.id} bucket")
    else:
        logger.info(
            f"The {file} is retrived successfully from the {usr_bucket.id} bucket"
        )

    return blob_usr


def glob(
    search_criteria: Union[str, List[str]], content: list, function: str = "startswith"
) -> List[str]:
    """gcs_filenames.

    Parameters
    ----------
    search_criteria: [Union[str, List[str]]]
        a string that exists in the name, or a list of separate strings that each one of them exists in a separate
        location in the name.
    content: [list]
        content of the gcp bucket.
    function:
        function to be used to search for the file in the bucket content list
        default si "startswith", "__contains__"
        >>> function = "startswith".

    Returns
    -------
    list
    """
    # in case there are spaces in the search criteria
    if isinstance(search_criteria, str) and len(search_criteria.split(" ")) > 1:
        prefix_parts = search_criteria.split(" ")
        search_criteria = prefix_parts[0]
        for i in range(1, len(prefix_parts)):
            search_criteria = search_criteria + "%20" + prefix_parts[i]

    fnames = []
    for i in range(len(content)):
        whole_name = content[i]
        # without the .tif at the end
        # if whole_name.__contains__("."):
        name_without_tif = whole_name.split(".")[0]
        # else:
        #     name_without_tif = whole_name
        # if the search_criteria is a string
        if isinstance(search_criteria, str):
            # if you changed the search criteria variable name you have to change it in the next line also and it
            # has to stay as a string for the compile and eval function to work.
            if eval(
                compile(
                    f"'{name_without_tif}'.{function}(search_criteria)",
                    "<string>",
                    "eval",
                )
            ):
                fnames.append(whole_name)
        elif isinstance(search_criteria, list):
            if all(x in whole_name for x in search_criteria):
                fnames.append(whole_name)

    fnames = [i.replace("%2F", "/") for i in fnames]
    fnames = [i.replace("%20", " ") for i in fnames]
    return fnames


def get_file_names(
    location_id: str = "",
    search_criteria: str = "",
    date: bool = False,
    content: list = [],
    function: str = "__contains__",
):
    """get_file_names.

        get_file_names

    Parameters
    ----------
    location_id : [str]
        if there is a location related word, inserted at the beginning of the file name.
    search_criteria : [str]
        the dataset id
    date : [bool]
        if there is a time stamp in the file name.
    content : [list]
        list containing the file names in GCS
    function : [str]
        function to be used to search for the files

    Returns
    -------
    list[str]
        list of file names
    """
    if location_id == "":
        original_name = search_criteria
    else:
        print(f"polygon -  {location_id}")
        original_name = f"{location_id}_{search_criteria}"

    name_prefix = original_name

    "the name_prefix if unique for each group of data"
    # name_prefix = f"{original_name}"

    # name_prefix = f"{original_name}%2F{dataset}%2F{original_name}_"
    # get the old files names based on the unique name_prefix
    old_fnames = glob(name_prefix, content, function=function)
    return old_fnames


def check_file_in_bucket(
    file: str,
    client: storage.client.Client,
    bucket_id: str,
    option: int = 2,
    project_id: str = "",
    search_function: str = "startswith",
) -> bool:
    """check_file_in_bucket.

        check_file_in_bucket checks if a fiven file exists inside a bucket or not

    Parameters
    ----------
    file: [str]
        file name in the form of root2/root2/aoi-name_dataset-name_YYYY_MM_DD.tif
        >>> file = "Montana_ls7_ndvi_2000_08_20.tif"
    client: [storage.client.Client]
        storage client you get when initializing the gcs
        >>> from unicloud.authenticate import create_client
        >>> your_service_key = "<you-service-account>"
        >>> client = create_client(your_service_key, project_id)
    bucket_id: [str]
        bucket id
        >>> bucket_id = "testing_repos"
    option: [int]
        - option = 1 :
                    if you want give the whole name of the file, then the function
                     will try to retrieve it from the bucket
        - option = 2 :
                    if you want to search with the file name prefix only and not the full
                    name, and also gives you criteria to search for the name prefix
    project_id: [str]
        project id, needed only incase of option=2
        >>> project_id = "gee-data-access"
    search_function: [str]
        the way to search for the file name prefix in the string, i.e "startswith"
        or "__contains__"
        >>> search_function = "startswith"
    Returns
    -------
        [True/False] whether a file exists in the bucket or not.
    """
    if option == 1:
        bucket = client.get_bucket(bucket_id)
        try:
            get_file_from_bucket(file, bucket)
            exist = True
        except ValueError:
            exist = False
    else:
        content = get_bucket_contents(client, bucket_id, project_id)
        old_fnames: List = glob(file, content, function=search_function)
        if len(old_fnames) >= 1:
            exist = True
        else:
            exist = False
    return exist


def upload_from_disk_to_gcs(
    file: str, folders: List[str], bucket: storage.bucket.Bucket
):
    """upload_from_disk_to_gcp.

        upload data from disk to gcs

    Parameters
    ----------
    file: [str]
        the file path on disk.
    folders: [list]
        rootdirectories in gcp, if these folders does not exit it will be created,
        the order of the folders is important,
        ex folders=[root1, root2]
        root1/
            root2
    bucket: [storage.bucket.Bucket]
        bucket object
        ex, bucket = storage.Bucket(bucket_name)

    Returns
    -------
    file will be uplaoded to the given bucket object
    """
    roots = [i + "/" for i in folders]
    rpath = ""
    for i in range(len(roots)):
        rpath += roots[i]

    fname = file.split("/")[-1]
    blob2 = bucket.blob(rpath + fname)

    blob2.upload_from_filename(filename=file)
    logger.info(f"The file is uploaded to the following path {rpath + fname}")


def download_from_gcs_to_disk(
    file: str, folders: List[str], bucket: storage.bucket.Bucket
):
    """download_from_gcs_to_disk.

        upload data from disk to gcs

    Parameters
    ----------
    file: [str]
        the file path on disk.
    folders: [list]
        root directories in gcp, if these folders does not exit it will be created,
        the order of the folders is important,
        ex folders=[root1, root2]
        root1/
            root2
    bucket: [storage.bucket.Bucket]
        bucket object
        ex, bucket = storage.Bucket(bucket_name)

    Returns
    -------
    file will be uplaoded to the given bucket object
    """
    # https://stackoverflow.com/questions/42555142/downloading-a-file-from-google-cloud-storage-inside-a-folder
    # TODO: not finished yet
    roots = [i + "/" for i in folders]
    rpath = ""
    for i in range(len(roots)):
        rpath += roots[i]

    fname = file.split("/")[-1]
    blob2 = bucket.blob(rpath + fname)

    blob2.upload_from_filename(filename=file)
    logger.info(f"The file is uploaded to the following path {rpath + fname}")


def move_files_to_folder(
    old_name: List[str],
    new_name: List[str],
    bucket_id: str,
    client_usr: storage.client.Client,
):
    """move_files_to_folder.

        move_files_to_folder renames/moves file in the same bucket into new names( inside folders)

    Parameters
    ----------
    old_name : list[str]
        list of the names of the files you want to move
    new_name : list[str]
        new file names including the whole path folder, if have to contain slash at the end ex ("fildername/")
    bucket_id: [str]
        bucket id
    client_usr: [client.client]
        client object from the google cloud storage module
        ex client = storage.Client(project=project_id, credentials=credentials)

    Returns
    -------
    True :[boolean]
        if the function finished the work successfully it will retun True

    Examples
    --------
    >>> old_name = ["modis_veg_20000101.tif"]
    >>> new_name = ["modis_veg_2000_01_01.tif"]
    >>> bucket_id = "testing-repos"
    >>> move_files_to_folder(old_name, new_name, bucket_id, client_usr)
    """
    # get the bucket you want
    usr_bucket = client_usr.get_bucket(bucket_id)
    # folder_name = "folder1/"
    for i in range(len(old_name)):
        logger.info(f"{i} / {len(old_name)} - {old_name[i]}")
        # create a blob for the file you want (file from the bucket)
        blobi = get_file_from_bucket(old_name[i], usr_bucket)
        # new_name = folder_name + files[i]
        usr_bucket.rename_blob(blobi, new_name[i])

    return True


def move_to_other_bucket(
    old_name: List[str],
    new_name: List[str],
    source_bucket_id: str,
    destination_bucket_id: str,
    client_usr: storage.client.Client,
    delete_old: bool = False,
):
    """move_to_other_bucket.

        the function does not check if a file with the same name already exists, but overwrite the old file if
        exists.
        # optimize: check if the file exists first in the destnation directory and warn the user

    Parameters
    ----------
    old_name : list[str]
        list of the names of the files you want to move
        >>> old_name = ["image_to_other_bucket.tif"]
    new_name : list[str]
        new file names including the whole path folder, if have to contain slash at the end ex ("fildername/")
        >>> new_name = ["root1/root2/image.tif"]
    source_bucket_id: [str]
        source bucket id
        >>> source_bucket_id = "testing_repos"
    destination_bucket_id: [str]
        destination bucket id
        >>> destination_bucket_id = "-datasets"
    client_usr: [client.client]
        client object from the google cloud storage module
        >>> from unicloud.authenticate import create_client
        >>> project_id = "gee-data-access"
        >>> your_service_key = <your_service_key>
        >>> client = create_client(your_service_key, project=project_id)
    delete_old : [bool]
        True if you want to delete the old files, True otherwise.
        >>> delete_old = False
    Returns
    -------
    None

    Examples
    --------
    >>> source_bucket_id = "testing_repos"
    >>> destination_bucket_id = "-datasets"
    >>> old_name = ["image_to_other_bucket.tif"]
    >>> client_usr = client
    >>> new_name = ["root3/image.tif"]
    """
    src_bucket = client_usr.get_bucket(source_bucket_id)
    dst_bucket = client_usr.get_bucket(destination_bucket_id)

    for i in range(len(old_name)):
        logger.info(f"{i} / {len(old_name)} - {old_name[i]}")
        # create a blob for the file you want (file from the bucket)
        src_blob = get_file_from_bucket(old_name[i], src_bucket)
        # new_name = folder_name + files[i]
        # src_bucket.rename_blob(source_blob, new_name[i])
        # copy to new destination
        src_bucket.copy_blob(src_blob, dst_bucket, new_name[i])
        if delete_old:
            # delete the old file
            src_blob.delete()


# TODO : create a function to rename a bucket


def delete_file(
    prefix: str,
    client: storage.client.Client,
    bucket: str,
    project_id: str,
    function: str = "startswith",
):
    """delete_file.

        delete_file deletes the files that their names resut from applying the function
        given to the function parameter

    Parameters
    ----------
    prefix : [str]
        prefix is any string that is included in the file name, it can be in the beginning, in the middle
        or in the end.
    client : [google.cloud.storage.client.Client]
        google client object
    bucket : [str]
        bucket id
    project_id : [str]
        project name
    function: [str]
        function to filter the files in the buckets. Default is "startswith"

    Returns
    -------
    None

    Examples
    --------
    >>> bucket = "-datasets"
    >>> prefix = "859932"
    >>> delete_file(prefix, client, bucket, project_id, function ="startswith")
    """
    contents = get_bucket_contents(client, bucket, project_id)
    files = glob(prefix, contents, function=function)
    bucket_usr = get_bucket(client, bucket, project_id)

    for i in range(len(files)):
        blob = get_file_from_bucket(files[i], bucket_usr)
        blob.delete()
        logger.info(f"File {files[i]} has been deleted")


def dataframe_to_file(df: pd.DataFrame, uri: str, gcp_client: storage.Client):
    """Store the DataFrame to a local or remote file."""
    buffer = io.BytesIO()
    df.to_csv(
        buffer,
        index=False,
        compression={
            "method": "gzip",
            "mtime": 1,  # For reproducibility: do not store timestamp.
        }
        if uri.endswith(".gz")
        else None,
    )
    buffer.seek(0)

    outfile = open_file(uri, mode="wb", gcp_client=gcp_client)
    shutil.copyfileobj(buffer, outfile)
    outfile.close()


def dataframe_from_file(uri: str, gcp_client: storage.Client):
    """Read a DataFrame from a local or remote file."""
    infile = open_file(uri, mode="rb", gcp_client=gcp_client)
    df = pd.read_csv(
        infile, index_col=False, compression="gzip" if uri.endswith(".gz") else None
    )
    infile.close()
    return df



""" This module is responsible for creating a GCS client """
from typing import Optional
from pathlib import Path
import os
import base64
import json
from unicloud.abstract_class import CloudStorageFactory
from google.oauth2 import service_account
from google.cloud import storage


class GCS(CloudStorageFactory):
    """GCS Cloud Storage"""

    def __init__(self, project_name: str, service_key: Optional[str] = None):
        """
        Initializes the GCS client.

        Parameters
        ----------
        project_name: [str]
            The Google Cloud project name.
        service_key: [str]
            The path to your service key file.
        """
        self._project_id = project_name
        if service_key is not None:
            if not Path(service_key).exists():
                raise FileNotFoundError(
                    f"The service key file {service_key} does not exist"
                )

        self.service_key = service_key
        self._client = self.create_client()

    @property
    def project_id(self):
        """project_id."""
        return self._project_id

    @property
    def client(self):
        """client."""
        return self._client

    def create_client(self) -> storage.client.Client:
        """create_client.

            the returned client deals with everything related to the specific project. For Google Cloud Storage,

            authenticating via a service account is the recommended approach. If you're running your code on a Google
            Cloud environment (e.g., Compute Engine, Cloud Run, etc.), the environment's default service account
            might automatically be used, provided it has the necessary permissions. Otherwise, you can set the
            GOOGLE_APPLICATION_CREDENTIALS environment variable to point to your service account JSON key file.

        Returns
        -------
        google cloud storage client object
        """
        if "GOOGLE_APPLICATION_CREDENTIALS" in os.environ:
            credentials = service_account.Credentials.from_service_account_file(
                os.environ["GOOGLE_APPLICATION_CREDENTIALS"]
            )
            client = storage.Client(project=self.project_id, credentials=credentials)
            # client = storage.Client(project=self.project_name)
        elif "SERVICE_KEY_CONTENT" in os.environ:
            # key need to be decoded into a dict/json object
            service_key_content = base64.b64decode(
                eval(f"b'{os.environ['SERVICE_KEY_CONTENT']}'")
            ).decode()

            service_key_content = json.loads(service_key_content)
            # connection to the service account
            client = storage.Client.from_service_account_info(service_key_content)
        elif self.service_key:
            credentials = service_account.Credentials.from_service_account_file(
                self.service_key
            )
            client = storage.Client(project=self.project_id, credentials=credentials)
        else:
            raise ValueError(
                "Since the GOOGLE_APPLICATION_CREDENTIALS and the EE_PRIVATE_KEY and EE_SERVICE_ACCOUNT are not in your"
                "env variables you have to provide a path to your service account"
            )

        return client

    def upload(self, file_path: str, destination: str):
        """Upload a file to GCS.

        Parameters
        ----------
        file_path: [str]
            The path to the file to upload.
        destination: [str]
            The destination path in the cloud storage.
        """
        bucket_name, object_name = destination.split("/", 1)
        bucket = self.client.bucket(bucket_name)
        blob = bucket.blob(object_name)
        blob.upload_from_filename(file_path)
        print(f"File {file_path} uploaded to {destination}.")

    def download(self, source, file_path):
        """Download a file from GCS.

        Parameters
        ----------
        source: [str]
            The source path in the cloud storage.
        file_path: [str]
            The path to save the downloaded file.
        """
        bucket_name, object_name = source.split("/", 1)
        bucket = self.client.bucket(bucket_name)
        blob = bucket.blob(object_name)
        blob.download_to_filename(file_path)
        print(f"File {source} downloaded to {file_path}.")
